import itertools, sys
import pandas as pd
import networkx as nx
from collections import Counter
from lib import backboning as bb
from pandarallel import pandarallel
from scipy.sparse import coo_matrix, csr_matrix

pandarallel.initialize(progress_bar = True)

def reconnect_singletons(df, df_bb, nodes):
   additional_edges = []
   missing_nodes = nodes - (set(df_bb["src"]) | set(df_bb["trg"]))
   for missing_node in missing_nodes:
      try:
         missing_node_edges = df[(df["src"] == missing_node) | (df["trg"] == missing_node)]
         additional_edges.append(pd.DataFrame(missing_node_edges.loc[missing_node_edges["score"].idxmax()]).T)
      except ValueError:
         pass
   return pd.concat([df_bb, pd.concat(additional_edges).drop_duplicates()])

def reconnect_components(df, df_bb):
   G = nx.from_pandas_edgelist(df_bb, source = "src", target = "trg")
   ccs = list(nx.connected_components(G))
   ccs = {n: i for i in range(len(ccs)) for n in ccs[i]}
   df["src_comp"] = df["src"].map(ccs)
   df["trg_comp"] = df["trg"].map(ccs)
   while df[["src_comp", "trg_comp"]].value_counts().size > 1:
      new_edge = df.loc[df.loc[df["src_comp"] != df["trg_comp"], "score"].idxmax()]
      df_bb = pd.concat([df_bb, pd.DataFrame(new_edge[["src", "trg", "nij", "score"]]).T])
      new_comp = new_edge[["src_comp", "trg_comp"]].min().min()
      old_comp = new_edge[["src_comp", "trg_comp"]].max().max()
      df.loc[df["src_comp"] == old_comp, "src_comp"] = new_comp
      df.loc[df["trg_comp"] == old_comp, "trg_comp"] = new_comp
   return df_bb

def artist_projection(df):
   sys.stderr.write("Counting artist co-occurrences...\n")
   co_occur = df.groupby(by = ["band", "year"])["artist"].apply(set)
   co_occur = Counter([y for x in co_occur.map(lambda x: list(itertools.combinations(x, 2))).values for y in x])
   co_occur = pd.DataFrame(data = [(min(pair[0], pair[1]), max(pair[0], pair[1]), co_occur[pair]) for pair in co_occur], columns = ("src", "trg", "nij"))
   sys.stderr.write("Backboning artist projection...\n")
   co_occur = co_occur.groupby(by = ["src", "trg"]).sum().reset_index()
   co_occur = pd.concat([co_occur, co_occur.rename(columns = {"src": "trg", "trg": "src"})])
   co_occur.index = range(co_occur.shape[0])
   co_occur_nc = bb.noise_corrected(co_occur, undirected = True, calculate_p_value = True)
   co_occur_nc_bb = bb.thresholding(co_occur_nc, 0.999999)
   sys.stderr.write("Reconnecting isolated nodes...\n")
   co_occur_nc_bb = reconnect_singletons(co_occur_nc, co_occur_nc_bb, set(df["artist"].unique()))
   sys.stderr.write("Reconnecting isolated components...\n")
   co_occur_nc_bb = reconnect_components(co_occur_nc, co_occur_nc_bb)
   sys.stderr.write("Writing artist projection...\n")
   co_occur_nc_bb[co_occur_nc_bb["src"] < co_occur_nc_bb["trg"]].to_csv("../outputs/projection_artist.txt", sep = "\t", index = False)

def band_projection(df):
   def artist_matcher(series, name):
      series[series.index < name] = series[series.index < name].index.map(lambda x: occur[name,:].minimum(occur[x,:]).sum())
      return series
   sys.stderr.write("Counting band overlaps...\n")
   occur = df.groupby(by = ["band", "artist"]).size().reset_index()
   _ = df["band"].drop_duplicates().sort_values().reset_index()
   band2row = _.reset_index().set_index("band")["level_0"].to_dict()
   row2band = _["band"].to_dict()
   _ = df["artist"].drop_duplicates().sort_values().reset_index()
   art2col = _.reset_index().set_index("artist")["level_0"].to_dict()
   occur["band"], occur["artist"] = occur["band"].map(band2row), occur["artist"].map(art2col)
   occur = csr_matrix(coo_matrix((occur[0], (occur["band"], occur["artist"]))))
   co_occur = pd.DataFrame(data = 0, index = range(occur.shape[0]), columns = range(occur.shape[0]))
   co_occur = co_occur.parallel_apply(lambda x: artist_matcher(x, x.name)).fillna(0)
   sys.stderr.write("Backboning band projection...\n")
   co_occur.index = co_occur.index.map(row2band)
   co_occur.columns = co_occur.columns.map(row2band)
   co_occur += co_occur.T
   co_occur = co_occur.unstack().reset_index()
   co_occur.columns = ("src", "trg", "nij")
   co_occur = co_occur[co_occur["nij"] > 0]
   band_net_nc = bb.noise_corrected(co_occur, undirected = True, calculate_p_value = True)
   band_net_nc_bb = bb.thresholding(band_net_nc, 0.9999998)
   band_net_nc_bb = band_net_nc_bb[band_net_nc_bb["src"] < band_net_nc_bb["trg"]]
   sys.stderr.write("Reconnecting isolated nodes...\n")
   band_net_nc_bb = reconnect_singletons(band_net_nc, band_net_nc_bb, set(df["band"].unique()))
   sys.stderr.write("Reconnecting isolated components...\n")
   band_net_nc_bb = reconnect_components(band_net_nc, band_net_nc_bb)
   sys.stderr.write("Writing band projection...\n")
   band_net_nc = band_net_nc[band_net_nc["src"] < band_net_nc["trg"]]
   band_net_nc[["src", "trg", "nij", "score"]].to_csv("../outputs/projection_band_full.txt", sep = "\t", index = False)
   band_net_nc_bb.to_csv("../outputs/projection_band.txt", sep = "\t", index = False)

sys.stderr.write("Processing input...\n")
df = pd.read_csv("../data/network.tsv", sep = "\t", dtype = {"band": str})
node_ids = pd.read_csv("../data/ids.tsv", sep = "\t", dtype = {"band": str, "id": str})

node_ids["band"] = node_ids["band"].str.strip()
node_ids["band"] = node_ids["band"].str.replace("’", "'")
df["artist"] = df["artist"].str.strip()
df["artist"] = df["artist"].str.replace("’", "'")
df["band"] = df["band"].str.strip()
df["band"] = df["band"].str.replace("’", "'")
df["band"] = df["band"].map(node_ids.set_index("band")["id"].to_dict())
df = df[["artist", "band", "year"]].drop_duplicates()

artist_projection(df)
band_projection(df)
